/***************************************************************
 ************************* A TPP Project ***********************
 ******************** Visit http://tpp.dhs.org *****************
 ***************************************************************
This file is a part of the: 

AHLL(AKOS High Level Language), pronounced: "all"

which is a part of the:

AKOS Project


This file is for internal USE only, it may NOT be distributed
outside the AKOS development team.

Author(s): Paul Fleischer (polle@denmark.crosswinds.net)
Created: 20. June 1999

Copyright(c) 1999 The AKOS team.
All Rights Reserved.
****************************************************************/

#include "main.h"

void internalFunc(CODE_BLOCK *code, char *name, int args, char in[10][256], char *out);

char InternalFunctions[10][255];
int NoOfFunctions;

//Test main function.
void
main(int args, char **argv)
{
  CODE_BLOCK testCode;
  char *codeString = "i=23 * 2\n\
print(\"Variable i: \")\n\
println(i)\n\
print(\"i * 2 + i is \")\n\
println(i*2+i)\n\
print(\"i * 2 + i == 5 (0 means false, and 1 means true): \")\n\
println(i*2+i == 5)\n\
l=i\n\
print(\"Variable i: \")\n\
println(i,\" \",\"3\")\n\
print(\"Variable l: \")\n\
println(l)\n\
println(\"Random number: \", random(5))\n\
.";

  printf("AHLL version %s\nCopyright (c) 1999 The Programming Place\n\n\n", VERSION);

  //Setup internal functions
  strcpy(InternalFunctions[0], "print");
  strcpy(InternalFunctions[1], "println");
  strcpy(InternalFunctions[2], "random");
  NoOfFunctions = 3;

  //Setup a test CODE_BLOCK
  strcpy(testCode.name, "test");
  testCode.linePointer = 0;
  testCode.data = (char*)malloc(strlen(codeString));
  strcpy(testCode.data, codeString);
  testCode.dataSize = strlen(testCode.data);
  getLines(&testCode);
  testCode.noOfVars = 0;

  //Now interpret the test CODE_BLOCK
  bool ret=true;

  for(int x=0;ret; x++)
    {
      dstring string;

      ret = nextLine(&testCode, &string);
      if(ret)
	{
	  #ifdef DEBUG
	  fprintf(stderr, "\nLine %i: %s\n", x, string.tochar());
	  #endif
	  /*	  removeSpaces(string);
	  if(translate(string, &testCode))
	    printf("Translated to: %s\n", string);
	  */
	  if(!analyseString(&string, &testCode))
	    printf("Ignoring line %i due to error(s)!\n", x+1);
	}
    }
  #ifdef DEBUG
  fprintf(stderr, "end reached\n");
  #endif
}


/**********************************************************************
-----------------------------------------------------------------------
void
getCommand(char *string, char *output)
-----------------------------------------------------------------------
Info: Returnes the command from a code string.

In: 'string'; Code string, which has been converted using translate.
    'output'; Pointer to where the name of the command should be put.

Out: 'output'; contains the command name on return.

Return: <NONE>
***********************************************************************/
void
getCommand(char *string, char *output)
{
  int i=0; //My very own counter variable

  //If the first character in the string is a space, then their might be more
  //spaces before the acutal command, and we'll have to fast forward ;)
  if(string[0] == ' ')
      for(;string[i] == ' ';i++);

  //Maybe the line is just a comment, and in that case ignore it!
  if(string[i] == '#') 
    {   
      strcpy(output, "Comment <NO ARGS>");
      return;
    }

  int x=0;
  for(; (string[i]!=' ') && (string[i]!='\0');i++,x++)
    output[x] = string[i];

  output[x] = '\0';
}



/**********************************************************************
-----------------------------------------------------------------------
void
removeSpaces(char *string)
-----------------------------------------------------------------------
Info: Removes all spaces, since they are unnecessary

In: 'string'; The string to be proccesed
    
Out: 'output'; contains the new string, with all spaces removed.

Return: void
***********************************************************************/
void
removeSpaces(char *string)
{
  //DAMN, the string is static... (has to be dynamic, to be real COOL ;)
  char output[255];
  int i=0; //my very own counter variable
  int l=0; //Now I have two counter variables ;)

  //Copy the string, but ignore spaces
  for(;string[i] != '\0';i++)
    {
      if(string[i] != ' ')
	{
	  output[l] = string[i];
	  l++;
	}
    }
  output[l] = 0;

  strcpy(string, output);
}


/**********************************************************************
-----------------------------------------------------------------------
void
emitError(char *module, int line, char *messgae)
-----------------------------------------------------------------------
Info: Sends error messages to the debugging system.
      Currently this function does only print a error message to stdout

In: 'module'; The name of module, were the error has occured.
    'line'; The line were the error has occured.
    'message'; The error message to send.
    
Out: <NONE>

Return: void
***********************************************************************/
void
emitError(char *module, int line, char *message)
{
  printf("\n*********************************\n");
  printf("PRE Monitoring System: ERROR:\n");
  printf("%s: line %i: %s\n", module, line, message);
  printf("*********************************\n");
}

/**********************************************************************
-----------------------------------------------------------------------
void
emitWarning(char *module, int line, char *messgae)
-----------------------------------------------------------------------
Info: Sends warning messages to the debugging system.
      Currently this function does only print a warning message to stderr

In: 'module'; The name of module, were the error has occured.
    'line'; The line were the error has occured.
    'message'; The error message to send.
    
Out: <NONE>

Return: void
***********************************************************************/
void
emitWarning(char *module, int line, char *message)
{
  fprintf(stderr, "\n****************************\n");
  fprintf(stderr, "PRE Monitoring System: WARNING:\n");
  fprintf(stderr, "%s: line %i: %s\n", module, line, message);
  fprintf(stderr, "****************************\n");
}

/**********************************************************************
-----------------------------------------------------------------------
bool
analyseString(dstring string, CODE_BLOCK *code)
-----------------------------------------------------------------------
Info: This function analyses the string, and decides what to do with 
      the different parts of it. The input string is unmodified on return

In:   'string'; the string to analyse, will soon be removed since 'code'
                can be used instead.
      'code';   the CODE_BLOCK which the string belongs to, it will very soon
                release the 'string' argument, and read the current line itself
    
Out: <NONE>

Return: bool, false if an error occured, true on success.
        Currently it is unknown if parts of the string have been executed
        on a false return value.
***********************************************************************/
bool
analyseString(dstring *string, CODE_BLOCK *code)
{
  int i=0; //Counter variable
  int l=0;
  char command[255];
  int start, end = 0; //Used to replace a function call with it's return value


  //First we need to ignore all whitespaces in the start of the string
  for(;(string->tochar()[i] == ' ') && (i<strlen(string->tochar()));i++);

  //If the character '"' is found as the first character, exit the function since we don't have to do anything with strings
  if(string->tochar()[i] == '"') return true;

  start = i; //Store the beginning of the command

  //Store the string until we incounter the following characters:
  //'(' start of an argument for a function call
  //'=' assigment
  //' '(whitespace) probably to make the code look better, if we encounter a whitespace, we have to stop the loop
  bool endLoop = false;
  for(; (string->tochar()[i] != '(') &&
	(string->tochar()[i] != '=') &&
	(i<strlen(string->tochar())) &&
	(!endLoop)
	;i++) 
    {
      if(string->tochar()[i] != ' ')
	{
	  command[l] = string->tochar()[i];
	  l++;
	} else endLoop = true;
    }
  
  command[l] = 0;

  if( (string->tochar()[i] == '=') && (string->tochar()[i+1] != '=') ) //Asignment
    {
      char value[255];
      bool isString;
      //Now we need to find the value to assign, but first jump all whitespaces

      i++; //We aren't interrested in the '=', we know that it's there ;)
      for(;string->tochar()[i] == ' ';i++);

      if(string->tochar()[i] == '"') isString = true; //The '"' tells us NOT to translate
                                            //the string with translate()
      //      if(isString) i++;
      //The rest of the string has to be the value
      for(l=0;i<strlen(string->tochar());i++,l++)
	value[l] = string->tochar()[i];

      value[l] = 0;
      //If our value isn't a string, we can remove spaces and calculate the result
      if(!isString)
	{
	  removeSpaces(value);
	  translate(value, code);
	}

      /*   
	   REMOVED
	   else
	value[l-1] = 0;	 //if it is a string, we'll have to remove the ending '"'
	REMOVED
      */


      //If the variable doesn't exist create it
      //We do not error checking, since it (hopefully) isn't necesarry
      if(!varExists(code, command))
	{
	  createVar(code, command);
	}

      //Finally we assign the variable the value 'value'
      assignVar(code, command, value);

      return true; //Return with success
    }
  else if(string->tochar()[i] == '(') //Execute/call a command/function
    {
      char args[10][256]; //DAMN... Too many statics in this programs ;)
      int argC =0;        //Argument Counter
      bool isString;
      int l=0;

      i++; //Well, we don't want the starting '(' to go into the first argument

      //First we need to find the arguments to the function
      //we have to keep in mind, that the number of arguments is unknown
      //to us right now(compile-time)

      //Arguments are seperated by a comma ','
      //Spaces are ignored (not copied to the argument string)
      //Parantheses(')') stop the loading of the argument list
      //Both spaces, commas and parantheses are ignored if we are inside a 
      //string, which is started and ended by a '"'
      //#############################################################
      //# The '"' character is only allowed as string starter ender #
      //# and it is NOT possible to overcome this by using '\"'     #
      //# this HAS to be changed VERY soon, but for now I can       #
      //# live without it ;)                                        #
      //#############################################################
      int pCounter;
      do
	{
	  isString = false;
	  pCounter=1;
	  l = 0;
	  for(; !( string->tochar()[i] == ',' && !isString) &&
		(i<strlen(string->tochar())) &&
		(pCounter > 0)
		;i++)
	    {
	      
	      if(string->tochar()[i] == '(') pCounter++;
	      if(string->tochar()[i] == ')') pCounter--;
	      
	      if(pCounter > 0)
		{
		  //if we are inside a string, do not remove spaces
		  if(isString)
		    {     
		      args[argC][l] = string->tochar()[i];
		      l++;	  
		    }	      
		  else
		    {
		      if(string->tochar()[i] != ' ')
			{
			  args[argC][l] = string->tochar()[i];
			  l++;
			}
		    }
		  
		  if(isString && string->tochar()[i] == '"')
		    isString = false;
		  else
		    if(!isString && string->tochar()[i] == '"')
		      isString = true;
		}
	    } //end of for-loop
	  i++; //to skip the ','

	  args[argC][l] = 0;
#ifdef DEBUG
	  fprintf(stderr, "%i. argument: %s\n", argC, args[argC]);
	  fprintf(stderr, "Now sending the argument through analyseString...\n");
#endif
	  end = i;
	  dstring argument(args[argC]);
	  analyseString(&argument, code);
	  strcpy(args[argC], argument.tochar());
#ifdef DEBUG
	  fprintf(stderr, "After analyseString(): '%s'\n", args[argC]);
#endif

	  argC++;
	} while(pCounter > 0);

      argC--;

      int count;

      //Normally we would check to see if the function is local, but again
      //for simplicity we only check if the function is global
      for(count=0;(count <= NoOfFunctions) && 
	    (strcmp(InternalFunctions[count], command) != 0)
	    ; count++);

      //If we found it in the InternalFunctions array, then call it!
      if(strcmp(InternalFunctions[count], command) == 0)
	{
	  char tmpstr[255];
	  char retVal[255];

	  internalFunc(code, command, argC+1, args, retVal);
	  strncpy(tmpstr, string->tochar(), start);
	  tmpstr[start] = 0;
	  strcat(tmpstr, retVal);
	  strcat(tmpstr, string->tochar()+end);

	  *string = tmpstr;
	}
      return true;
    }

}


/**********************************************************************
-----------------------------------------------------------------------
void 
internalFunc(CODE_BLOCK *code, char *name, int args, char *in, char *out)
-----------------------------------------------------------------------
Info: This function executes internal functions provided by hll.
      Currently it is only temporarely, since it later will be possible to
      use c++ functions as internal AHHL functions.
      Only one argument can be passed to the internal functions at this stage.

In: code: the CODE_BLOCK
    name: the function name to execute
    args: Number of arguments, must always be 1
    in: argument for the functions
        
Out: out: Return value of the function

Return: void
***********************************************************************/
void 
internalFunc(CODE_BLOCK *code, char *name, int args, char in[10][256], char *out)
{
  //Take care of internal functions

  //print() has one argument, futher agument will be ignored and a warning
  //will be send to the monitoring system
  if( (strcmp(name, "print") == 0) || (strcmp(name, "println") == 0) )
    {
      if(args < 1)
	{
	  emitWarning(code->name, code->linePointer, "Too few arguments for print(), minimum 1 argument required!");
	  return;
	}

#ifdef DEBUG
      fprintf(stderr, "print(ln) DEBUG START\n");
#endif

      for(int i=0; i<args;i++)
	{
#ifdef DEBUG
	  fprintf(stderr, "%i. arg: '%s'\n", i, in[i]);
	  fprintf(stderr, "in[i][0]: %c\n", in[i][0]);
#endif
	  if(in[i][0] != 0)
	    {
	      if(in[i][0] != '"')
		{
		  removeSpaces(in[i]);
		  translate(in[i], code);
		  printf("%s", in[i]);
		}
	      else
		{
		  char tmpstr[255];
		  strcpy(tmpstr, in[i]+1);
		  tmpstr[strlen(tmpstr)-1] = 0;
		  printf("%s", tmpstr);
		}
	    }  
	}
#ifdef DEBUG
      fprintf(stderr, "print(ln) DEBUG END\n");
#endif
      if(strcmp(name, "println") == 0)
	printf("\n");
     
      return;
    }
  
  if(strcmp(name, "random") == 0)
    {
      if( (args != 1) || (in[0] == 0))
	{
	  emitWarning(code->name, code->linePointer, "random requires 1 argument");
	  out[0] = 0;
	  return;
	}
      unsigned int tmpInt = atoi(in[0]);


      tmpInt = 1+(int) (tmpInt*rand()/(RAND_MAX+1.0));

      sprintf(out, "%i", tmpInt); 
      
      return;
    }
}



